// EXPORTS.CPP
// WaveFront Gatekeeper exported functions
// In most cases, these are simply wrapper functions that call
// the appropriate function in the specified CSynth. All functions are
// protected to deal well with out-of-range synth indices (which will most
// likely only happen if there is no Wavefront synth present).
// 2/15/94 J. Johnson
// Copyright  1994 Turtle Beach Systems, Inc. All Rights Reserved

#include <afxwin.h>
#include <afxdlgs.h>     
#include <math.h>
#include <mmsystem.h>
#include <time.h>
#include <fstream.h>
#include <mmreg.h>       // multimedia format defines
#include "wfssxapi.h"
#include "wfsynth.h"
#include "wfgatapi.h"
#include "prodids.h"
#include "wfsysex.h"
#include "midimsg.h"
#include "midiize.h"
#include "resource.h"
#include "wave.h"
#include "files.h"
#include "synth.h"
#include "wfgate.h"

extern CWaveFrontGatekeeperDLL NEAR Gatekeeper;

////////////////////////////////////////////////////////////////////////
// Ownership tracking functions

// Set the owner of the Gatekeeper - that is, the task that is currently 
// using the DLL. This should be set anytime an app that uses the DLL 
// is activated, and also checked when an app is activated. If some other
// app has accessed the DLL, then the app should get all data from the 
// Gatekeeper.    
// ** This should be the very first call made to the Gatekeeper. **
void FAR PASCAL _export GateSetOwner(HTASK hTask, HWND hWindow)
{
    Gatekeeper.m_hOwnerTask = hTask;  
    Gatekeeper.m_hOwnerWindow = hWindow;   
    
    // Clean up the list of clients by removing any task handles that
    // are no longer valid.
    for (int i = 0; i < MAX_CLIENTS; i++)
    {
        if (Gatekeeper.m_hClientTask[i] != NULL && !IsTask(Gatekeeper.m_hClientTask[i]))
        {
            Gatekeeper.m_hClientTask[i] = NULL;
            
            // Also set all dirty flags for this task
            for (UINT nSynth = 0; nSynth < Gatekeeper.m_nSynthCount; nSynth++)
            {
                for (int j = 0; j < NUM_PROGRAMS; j++)
                    Gatekeeper.m_synthArray[nSynth]->SetProgramDirtyFlag(j, i, TRUE);
                for (j = 0; j < NUM_PATCHES; j++)
                    Gatekeeper.m_synthArray[nSynth]->SetPatchDirtyFlag(j, i, TRUE);
                for (j = 0; j < NUM_SAMPLES; j++)
                    Gatekeeper.m_synthArray[nSynth]->SetSampleDirtyFlag(j, i, TRUE);
            }
        }
    }
    
    // Add this task to the list of clients.
    for (i = 0; i < MAX_CLIENTS; i++)
    {
        if (Gatekeeper.m_hClientTask[i] == NULL || Gatekeeper.m_hClientTask[i] == hTask)
        {
            Gatekeeper.m_hClientTask[i] = hTask;
            Gatekeeper.m_nCurrentOwner = i;
            break;
        }
    }                  
    ASSERT(i < MAX_CLIENTS);
}  

// Get the handle of the task that "owns" (is currently using) the DLL
HTASK FAR PASCAL _export GateGetOwner()
{                   
    return(Gatekeeper.m_hOwnerTask);
}

// Get the main window of the task that "owns" (is currently using) the DLL
HWND FAR PASCAL _export GateGetOwnerWindow()
{                   
    return(Gatekeeper.m_hOwnerWindow);
}

// See how many applications are using the DLL.
int FAR PASCAL _export GateGetReferenceCount()
{
    int nResult = GetModuleUsage(AfxGetInstanceHandle());
    TRACE("Gatekeeper reference count = %d\n", nResult);
    return(nResult);
}          

///////////////////////////////////////////////////////////////
// Gatekeeper initialization functions
// 

// Search for WaveFront synths and enumerate them.
// ** This should be the second call made to the Gatekeeper **
void FAR PASCAL _export GateInitialize()
{
    if (!Gatekeeper.m_bAreSynthsCounted)
        Gatekeeper.FindWaveFrontSynths();
}

//////////////////////////////////////////////////////////////////
// Synth informational functions

// Return the number of WaveFront synths in the system.
UINT FAR PASCAL _export GateGetWaveFrontSynthCount()
{
    if (!Gatekeeper.m_bAreSynthsCounted)
        GateInitialize();
    return(Gatekeeper.m_nSynthCount);
}

// Get the capabilities of a particular WaveFront synth output. Note that
// nSynth refers to the nth *WaveFront* synth, NOT the nth MIDI device!
void FAR PASCAL _export GateGetWaveFrontSynthOutputCaps(UINT nSynth, LPMIDIOUTCAPS lpCaps)
{
    if (!Gatekeeper.m_bAreSynthsCounted)
        GateInitialize();
    midiOutGetDevCaps(Gatekeeper.m_synthArray[nSynth]->GetOutputID(), lpCaps, 
        sizeof(MIDIOUTCAPS));
}

// Get the capabilities of a particular WaveFront synth input. Note that
// nSynth refers to the nth *WaveFront* synth, NOT the nth MIDI device!
void FAR PASCAL _export GateGetWaveFrontSynthInputCaps(UINT nSynth, LPMIDIINCAPS lpCaps)
{
    if (!Gatekeeper.m_bAreSynthsCounted)
        GateInitialize();
        
    // If the synth is undetectable, fill out a bogus MIDIINCAPS
    if (Gatekeeper.m_synthArray[nSynth]->IsUndetectable())
    {   
        CString strUndetectable;
        strUndetectable.LoadString(IDS_UNDETECTABLE);
        memset(lpCaps, 0, sizeof(MIDIINCAPS));
        lstrcpy(lpCaps->szPname, strUndetectable);
    }
    else
        midiInGetDevCaps(Gatekeeper.m_synthArray[nSynth]->GetInputID(), lpCaps, 
            sizeof(MIDIINCAPS));
}
 
// ask the synth if it has effects
BOOL FAR PASCAL _export GateDoesSynthHaveEffects(UINT nSynth)
{
    if (!Gatekeeper.m_bAreSynthsCounted)
        GateInitialize();
    if (nSynth < Gatekeeper.m_nSynthCount)
        return(Gatekeeper.m_synthArray[nSynth]->HasEffects());
    else
        return(FALSE);
}
                                  
// ask the synth if it has filters
BOOL FAR PASCAL _export GateDoesSynthHaveFilters(UINT nSynth)
{
    if (!Gatekeeper.m_bAreSynthsCounted)
        GateInitialize();
    if (nSynth < Gatekeeper.m_nSynthCount)
        return(Gatekeeper.m_synthArray[nSynth]->HasFilters());
    else
        return(FALSE);
}       

// Ask the synth if it is detectable (that is, if it will send
// sysex messages back to the system).
BOOL FAR PASCAL _export GateIsSynthUndetectable(UINT nSynth)
{
    if (!Gatekeeper.m_bAreSynthsCounted)
        GateInitialize();
    if (nSynth < Gatekeeper.m_nSynthCount)
        return(Gatekeeper.m_synthArray[nSynth]->IsUndetectable());
    else
        return(TRUE);
}

// Get the name of the synth's patch & sample name file (eg "MAUI.INI")
void FAR PASCAL _export GateGetSynthIniFileName(UINT nSynth, LPSTR lpszFileName)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
        lstrcpy(lpszFileName, Gatekeeper.m_synthArray[nSynth]->m_strNameFile);
    else
        lpszFileName[0] = 0;
}

// Get the name of the specified synth. This will be the synth's
// model name, optionally followed by a number if there are multiple
// synths of the same type present.
void FAR PASCAL _export GateGetSynthName(UINT nSynth, LPSTR lpszName)
{                                                               
    CString strName;

    if (nSynth < Gatekeeper.m_nSynthCount)
        strName = Gatekeeper.m_synthArray[nSynth]->GetName();
    else
        strName.Empty();
    
    lstrcpy(lpszName, strName);
}

// Get the range of the "Max Voices" setting. This may vary for future chipsets.
void FAR PASCAL _export GateGetMinMaxVoices(UINT nSynth, LPINT lpMin, LPINT lpMax)
{
    ASSERT(lpMin != NULL);
    ASSERT(lpMax != NULL);
    
    if (nSynth < Gatekeeper.m_nSynthCount)
    {                                                      
        *lpMin = Gatekeeper.m_synthArray[nSynth]->GetMinVoices();
        *lpMax = Gatekeeper.m_synthArray[nSynth]->GetMaxVoices();
    }   
    else
    {              
        // Values for the Maui
        *lpMin = 8;
        *lpMax = 32;
    }
}

///////////////////////////////////////////////////////////////////////
// Synth initialization functions

// Initialize the specified synthesizer. This opens its ports, and initializes 
// the data either from a file or the hardware. It also gets some data from
// the INI file (WFPATCH.INI).
HMIDIOUT FAR PASCAL _export GateInitializeSynth(UINT nSynth, LPCSTR pszFilespec)
{
    HMIDIOUT hResult = NULL;
    
    // First ensure that the Gatekeeper itself has
    // found all synths, and do so if not.
    if (!Gatekeeper.m_bAreSynthsCounted)
        GateInitialize();
    
    if (nSynth < Gatekeeper.m_nSynthCount)
    {                                                           
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        if (Gatekeeper.m_synthArray[nSynth]->Initialize(pszFilespec))
            hResult = Gatekeeper.m_synthArray[nSynth]->GetMidiOut();
        Gatekeeper.m_bIsBusy = FALSE;
    }

    return(hResult);
}

// Find out if the synth has been properly initialized.
BOOL FAR PASCAL _export GateIsSynthInitialized(UINT nSynth)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
        return(Gatekeeper.m_synthArray[nSynth]->IsInitialized());
    else
        return(FALSE);
}                                      

// Open a synth's MIDI ports. This generally only needs to be done if the
// ports were specifically closed. This function can also be called to
// get the handle to a previously opened MIDI output.
HMIDIOUT FAR PASCAL _export GateOpenSynth(UINT nSynth)
{
    HMIDIOUT hResult = NULL;
    
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        if (Gatekeeper.m_synthArray[nSynth]->OpenSynth())
            hResult = Gatekeeper.m_synthArray[nSynth]->GetMidiOut();
        Gatekeeper.m_bIsBusy = FALSE;
    }

    return(hResult);
}

// Close a synth's MIDI ports and save its master settings to disk.
// This MUST be called when an app that uses the Gatekeeper shuts down!
void FAR PASCAL _export GateCloseSynth(UINT nSynth)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        // Save all master settings to the INI file.
        // This has to be done here rather than within CSynth::CloseSynth because
        // CloseSynth is called from CSynth's destructor, at 
        // which time it seems to be impossible to access the synth.
        CString strSection = Gatekeeper.m_synthArray[nSynth]->GetName();
        CString strEntry;
        
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        // Open the synth if necessary (possibly necessary for SB16 drivers, etc.) 
        Gatekeeper.m_synthArray[nSynth]->OpenSynth();
        
        // .. tuning
        strEntry.LoadString(IDS_TUNING);
        AfxGetApp()->WriteProfileInt(strSection, strEntry, 
            Gatekeeper.m_synthArray[nSynth]->GetTuning());
        // .. volume
        strEntry.LoadString(IDS_MASTERVOLUME);
        AfxGetApp()->WriteProfileInt(strSection, strEntry, 
            Gatekeeper.m_synthArray[nSynth]->GetSynthVolume());
        // .. max voices
        strEntry.LoadString(IDS_MAXVOICES);
        AfxGetApp()->WriteProfileInt(strSection, strEntry, 
            Gatekeeper.m_synthArray[nSynth]->GetNumberOfVoices());
        // .. drum channel (trickier)
        int nDrumChannel = 0; 
        for (int i = 1; i <= MAX_MIDICHANNEL; i++)
        {
            if (Gatekeeper.m_synthArray[nSynth]->GetDrumChannelStatus(i))
            {
                nDrumChannel = i;
                break;
            }
        }     

        strEntry.LoadString(IDS_DRUMCHANNEL);
        AfxGetApp()->WriteProfileInt(strSection, strEntry, nDrumChannel);
            
        // Now close the MIDI ports    
        Gatekeeper.m_synthArray[nSynth]->CloseSynth();

        Gatekeeper.m_bIsBusy = FALSE;
    }
}

/////////////////////////////////////////////////////////////////////////
// "Dirty" tracking functions. The Gatekeeper keeps track of accesses to 
// patch, program, and sample data (but not drum kit, global, or effects data)
// for each client app. If an app does not know about changes to a particular
// element, that element is marked dirty.

// Check the dirty state of a sample for the current owner.
BOOL FAR PASCAL _export GateIsSampleDirty(UINT nSynth, int nSample)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
        return(Gatekeeper.m_synthArray[nSynth]->IsSampleDirty(nSample, 
            Gatekeeper.m_nCurrentOwner));
    else
        return(FALSE);
}                                      

// Check the dirty state of a program for the current owner.
BOOL FAR PASCAL _export GateIsProgramDirty(UINT nSynth, int nProgram)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
        return(Gatekeeper.m_synthArray[nSynth]->IsProgramDirty(nProgram, 
            Gatekeeper.m_nCurrentOwner));
    else
        return(FALSE);
}                                      

// Check the dirty state of a patch for the current owner.
BOOL FAR PASCAL _export GateIsPatchDirty(UINT nSynth, int nPatch)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
        return(Gatekeeper.m_synthArray[nSynth]->IsPatchDirty(nPatch, 
            Gatekeeper.m_nCurrentOwner));
    else
        return(FALSE);
}

/////////////////////////////////////////////////////////////////////
// Reset functions.                                      

// Reset the entire synth, wiping out its memory. Use sparingly.
void FAR PASCAL _export GateResetSynth(UINT nSynth)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());

        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->Reset();
        Gatekeeper.m_synthArray[nSynth]->GetItAll();

        Gatekeeper.m_bIsBusy = FALSE;
    }
}

/////////////////////////////////////////////////////////////////////////
// Audition functions

// Select an audition input for the system. The input is the device ID for the
// input to use. The function returns TRUE on success, FALSE on failure.
BOOL FAR PASCAL _export GateSelectAuditionInput(UINT nInputID)
{                                               
    return(Gatekeeper.SelectAuditionInput(nInputID));
}   

// Select an audition output for the system. The output is the device ID for the
// output to use. The function returns TRUE on success, FALSE on failure.
void FAR PASCAL _export GateSelectAuditionOutput(UINT nSynth)
{                                               
    if (nSynth < Gatekeeper.m_nSynthCount)
        Gatekeeper.SelectAuditionOutput(nSynth);
}   

// Enable or disable audition. This especially intended for times when sysex 
// data is being retrieved from the synth. Audition is on by default.
// This function simply calls a function in WFSYSEX.DLL so external apps
// don't need to know about that DLL.
void FAR PASCAL _export GateEnableAudition(BOOL bEnable) 
{
    EnableAudition(bEnable);
}                                                          

// Tell the MIDI input handler to rechannelize data. The input value is
// a true MIDI channel (1 to 16); a value of 0 disables rechannelization.
// This function simply calls a function in WFSYSEX.DLL so external apps
// don't need to know about that DLL.
void FAR PASCAL _export GateSetRechannelizeChannel(int nRechannelizeChannel)
{
    SetRechannelizeChannel(nRechannelizeChannel);
}                  

//////////////////////////////////////////////////////////////////////
// Set and get global parameters

// Set the specified synthesizer's volume
void FAR PASCAL _export GateSetSynthVolume(UINT nSynth, int nVolume)
{                                                            
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->SetSynthVolume(nVolume);
        Gatekeeper.m_bIsBusy = FALSE;
    }
}

// Get the specified synthesizer's volume
int FAR PASCAL _export GateGetSynthVolume(UINT nSynth)
{   
    int nVolume = 0;
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        nVolume = Gatekeeper.m_synthArray[nSynth]->GetSynthVolume();
        Gatekeeper.m_bIsBusy = FALSE;
    }               
    
    return(nVolume);
}

// Set the synthesizer's maximum voice count
void FAR PASCAL _export GateSetNumberOfVoices(UINT nSynth, int nVoices)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->SetNumberOfVoices(nVoices);

        Gatekeeper.m_bIsBusy = FALSE;
    }
}

// Get the synthesizer's maximum voice count
int FAR PASCAL _export GateGetNumberOfVoices(UINT nSynth)
{   
    int nResult = 24;  // This value should be in range for all variants 
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        nResult = Gatekeeper.m_synthArray[nSynth]->GetNumberOfVoices();

        Gatekeeper.m_bIsBusy = FALSE;
    }

    return(nResult);
}

// Set the specified synthesizer's tuning value.
// Tuning is in units of 1/2048 octave.
void FAR PASCAL _export GateSetTuning(UINT nSynth, int nTuning)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->SetTuning(nTuning);

        Gatekeeper.m_bIsBusy = FALSE;
    }
}

// Get the specified synthesizer's tuning value. 
// Tuning is in units of 1/2048 octave.
int FAR PASCAL _export GateGetTuning(UINT nSynth)
{
    int nTuning = 0;
    
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        nTuning = Gatekeeper.m_synthArray[nSynth]->GetTuning();
        Gatekeeper.m_bIsBusy = FALSE;
    }        
    
    return(nTuning);
}

// Set the specified channel to Drum Program or Normal mode.
// The channel is a true MIDI channel number (1 - 16)
void FAR PASCAL _export GateSetDrumChannelStatus(UINT nSynth, int nChannel, 
    BOOL bStatus)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->SetDrumChannelStatus(nChannel, bStatus);

        Gatekeeper.m_bIsBusy = FALSE;
    }        
}

// See if the specified channel is a drum channel.
// The channel is a true MIDI channel number (1 - 16)
BOOL FAR PASCAL _export GateGetDrumChannelStatus(UINT nSynth, int nChannel)
{                  
    BOOL bResult = (nChannel == 10);   // return TRUE for the default drum channel
    
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        bResult = Gatekeeper.m_synthArray[nSynth]->GetDrumChannelStatus(nChannel);
        Gatekeeper.m_bIsBusy = FALSE;
    }        
    return(bResult);
}

//////////////////////////////////////////////////////////////////////////
// WaveFront Synth system info ("report") functions

// Return the firmware version. The minor version is in the low byte,
// major in the high. If, for some reason, the synth does not respond,
// return an error code (0xFFFF)
WORD FAR PASCAL _export GateReportFirmwareVersion(UINT nSynth)
{
    WORD wVersion = 0;

    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        wVersion = Gatekeeper.m_synthArray[nSynth]->ReportFirmwareVersion();
        
        Gatekeeper.m_bIsBusy = FALSE;
    }        

    return(wVersion);
}

// Return the hardware version. The minor version is in the low byte,
// major in the high.
WORD FAR PASCAL _export GateReportHardwareVersion(UINT nSynth)
{
    WORD wVersion = 0;

    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        wVersion = Gatekeeper.m_synthArray[nSynth]->ReportHardwareVersion();

        Gatekeeper.m_bIsBusy = FALSE;
    }        

    return(wVersion);
}

// Get the amount of RAM and ROM in the system.
void FAR PASCAL _export GateReportConfiguration(UINT nSynth, LPDWORD pdwTotalRam, 
    LPDWORD pdwTotalRom)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->ReportConfiguration(pdwTotalRam, pdwTotalRom);

        Gatekeeper.m_bIsBusy = FALSE;
    }        
    else
    {                                                                   
        *pdwTotalRam = 0;
        *pdwTotalRom = 0;
    }
}

// Report the amount of free memory in the synth, in bytes.
// This also causes memory to be compacted.
DWORD FAR PASCAL _export GateReportFreeMemory(UINT nSynth)
{
    DWORD dwResult = 0;
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        dwResult = Gatekeeper.m_synthArray[nSynth]->ReportFreeMemory();

        Gatekeeper.m_bIsBusy = FALSE;
    }        
    
    return(dwResult);
}

// Report the number of samples slots in use, including internal and
// external samples, multisamples, etc.
WORD FAR PASCAL _export GateReportNumberOfSamples(UINT nSynth)
{
    WORD wResult = 0;

    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        wResult = Gatekeeper.m_synthArray[nSynth]->ReportNumberOfSamples();

        Gatekeeper.m_bIsBusy = FALSE;
    }        

    return(wResult);
}

// Report the program assigned to each channel. The CHANNEL_PROGRAMS struct holds
// standard MIDI program numbers, or PR_NORMALDRUM or PR_ENHANCEDDRUM for drum programs.
// (PR_NORMALDRUM is unused by this implementatin of the Gatekeeper; all drum kits
// are "enhanced" drum programs.)
void FAR PASCAL _export GateReportChannelPrograms(UINT nSynth, 
    LPCHANNEL_PROGRAMS lpChannelPrograms)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->ReportChannelPrograms(lpChannelPrograms);

        Gatekeeper.m_bIsBusy = FALSE;
    }        
    else
    {
        for (int i = 0; i < NUM_MIDICHANNELS; i++)
            lpChannelPrograms->byProgram[i] = 0;
    }
}

// Report the instantaneous output levels on both channels.
void FAR PASCAL _export GateReportLevels(UINT nSynth, LPINT lpLeftLevel, 
    LPINT lpRightLevel)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->ReportLevels(lpLeftLevel, lpRightLevel);

        Gatekeeper.m_bIsBusy = FALSE;
    }        
    else
    {                                                                   
        *lpLeftLevel  = 0;
        *lpRightLevel = 0;
    }
}

// Report the peak output levels on both channels.
// The return value is the saturation count. (See the Mockingbird docs for details.)
int FAR PASCAL _export GateReportPeakLevels(UINT nSynth, LPINT lpLeftLevel, 
    LPINT lpRightLevel)
{
    int nResult = 0;

    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        nResult = Gatekeeper.m_synthArray[nSynth]->ReportPeakLevels(lpLeftLevel, 
            lpRightLevel);

        Gatekeeper.m_bIsBusy = FALSE;
    }        
    else
    {                                                                   
        *lpLeftLevel  = 0;
        *lpRightLevel = 0;
    }

    return(nResult);
}

//////////////////////////////////////////////////
// Send/receive programs
             
// Stuff a program into the synthesizer, and save its name
void FAR PASCAL _export GateSetProgram(UINT nSynth, int nProgram, 
    LPPROGRAM lpProgram, LPCSTR pszName)
{                                                                                                     
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           

        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->SetProgram(nProgram, lpProgram, pszName);
        Gatekeeper.m_synthArray[nSynth]->CalculatePatchUsages();
        Gatekeeper.m_synthArray[nSynth]->CalculateSampleUsages();

        Gatekeeper.m_bIsBusy = FALSE;        
        
        // Set the dirty flags for all clients except the current
        for (int i = 0; i < MAX_CLIENTS; i++)
        {                                                                          
            BOOL bDirty = (i != Gatekeeper.m_nCurrentOwner);
            Gatekeeper.m_synthArray[nSynth]->SetProgramDirtyFlag(nProgram, i, bDirty);
        }
    }
}

// Get the specified program and program name from the synth
void FAR PASCAL _export GateGetProgram(UINT nSynth, int nProgram, LPPROGRAM lpProgram, 
    LPSTR pszName)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           

        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->GetProgram(nProgram, lpProgram, pszName);

        Gatekeeper.m_bIsBusy = FALSE;
        
        // Clear the dirty flag for the current owner
        Gatekeeper.m_synthArray[nSynth]->SetProgramDirtyFlag(nProgram, 
            Gatekeeper.m_nCurrentOwner, FALSE);
    }
    else
    {
        CString strProgram;
        
        strProgram.LoadString(IDS_PROGRAM);     
        wsprintf(pszName, "%s %d", (LPCSTR)strProgram, nProgram);
        memset(lpProgram, 0, sizeof(PROGRAM));
    }
}

// Set the name of a program
void FAR PASCAL _export GateSetProgramName(UINT nSynth, int nProgram, LPCSTR pszName)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           

        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->SetProgramName(nProgram, pszName);

        Gatekeeper.m_bIsBusy = FALSE;

        // Set the dirty flags for all clients except the current
        for (int i = 0; i < MAX_CLIENTS; i++)
        {                                                                          
            BOOL bDirty = (i != Gatekeeper.m_nCurrentOwner);
            Gatekeeper.m_synthArray[nSynth]->SetProgramDirtyFlag(nProgram, i, bDirty);
        }
    }
}

// Get the name of a program
void FAR PASCAL _export GateGetProgramName(UINT nSynth, int nProgram, LPSTR pszName)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           
        Gatekeeper.m_synthArray[nSynth]->GetProgramName(nProgram, pszName);

        // Clear the dirty flag for the current owner
        Gatekeeper.m_synthArray[nSynth]->SetProgramDirtyFlag(nProgram, 
            Gatekeeper.m_nCurrentOwner, FALSE);
    }
    else
    {
        CString strProgram;
        
        strProgram.LoadString(IDS_PROGRAM);     
        wsprintf(pszName, "%s %d", (LPCSTR)strProgram, nProgram);
    }
}   

////////////////////////////////////////////////////////////////////////////////
// Patch informational functions

// Find out how many times the specified patch is used.
int FAR PASCAL _export GateGetPatchUsageCount(UINT nSynth, int nPatch)
{                                               
    if (nSynth < Gatekeeper.m_nSynthCount)
        return(Gatekeeper.m_synthArray[nSynth]->GetPatchUsageCount(nPatch));
    else
        return(0);
}

// Locate an unused patch. nFirstPatch is the first patch that will be tested.
// If an unused patch is not found above nFirstPatch, the program will
// search below nFirstPatch. The return value is NO_PATCH (-1) if none is found.
// Note that a patch remains unused until a program or drumkit that uses the patch
// is actually sent to the Gatekeeper; to find additional unused slots, 
// call this function and pass in the number of the first patch following the last
// unused patch found.
int FAR PASCAL _export GateFindUnusedPatch(UINT nSynth, int nFirstPatch)
{    
    int nUnusedPatch = NO_PATCH;
    
    if (nSynth < Gatekeeper.m_nSynthCount)   
        nUnusedPatch = Gatekeeper.m_synthArray[nSynth]->FindUnusedPatch(nFirstPatch); 

    TRACE("Unused patch = %d\n", nUnusedPatch);
    return(nUnusedPatch);
}         

/////////////////////////////////////////////////////////////////////////
// Send/receive patches

// Stuff a patch into the synthesizer, and save its name
void FAR PASCAL _export GateSetPatch(UINT nSynth, int nPatch, LPPATCH lpPatch, 
    LPCSTR pszName)
{                                                                                                     
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           

        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->SetPatch(nPatch, lpPatch, pszName);
        Gatekeeper.m_synthArray[nSynth]->CalculateSampleUsages();

        Gatekeeper.m_bIsBusy = FALSE;
        
        // Set the dirty flags for all clients except the current
        for (int i = 0; i < MAX_CLIENTS; i++)
        {                                                                          
            BOOL bDirty = (i != Gatekeeper.m_nCurrentOwner);
            Gatekeeper.m_synthArray[nSynth]->SetPatchDirtyFlag(nPatch, i, bDirty);
        }
    }
}

// Get the specified patch and patch name from the synth
void FAR PASCAL _export GateGetPatch(UINT nSynth, int nPatch, LPPATCH lpPatch, 
    LPSTR pszName)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           

        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->GetPatch(nPatch, lpPatch, pszName);

        Gatekeeper.m_bIsBusy = FALSE;
        
        // Clear the dirty flag for the current owner
        Gatekeeper.m_synthArray[nSynth]->SetPatchDirtyFlag(nPatch, 
            Gatekeeper.m_nCurrentOwner, FALSE);
    }
    else
    {
        CString strPatch;
        
        strPatch.LoadString(IDS_PATCH);     
        wsprintf(pszName, "%s %d", (LPCSTR)strPatch, nPatch);
        memset(lpPatch, 0, sizeof(PATCH));
    }
}

// Set the name of a patch
void FAR PASCAL _export GateSetPatchName(UINT nSynth, int nPatch, LPCSTR pszName)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           

        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;

        Gatekeeper.m_synthArray[nSynth]->SetPatchName(nPatch, pszName);

        Gatekeeper.m_bIsBusy = FALSE;
        
        // Set the dirty flags for all clients except the current
        for (int i = 0; i < MAX_CLIENTS; i++)
        {                                                                          
            BOOL bDirty = (i != Gatekeeper.m_nCurrentOwner);
            Gatekeeper.m_synthArray[nSynth]->SetPatchDirtyFlag(nPatch, i, bDirty);
        }
    }
}

// Get the name of a patch
void FAR PASCAL _export GateGetPatchName(UINT nSynth, int nPatch, LPSTR pszName)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           
        Gatekeeper.m_synthArray[nSynth]->GetPatchName(nPatch, pszName);

        // Clear the dirty flag for the current owner
        Gatekeeper.m_synthArray[nSynth]->SetPatchDirtyFlag(nPatch, 
            Gatekeeper.m_nCurrentOwner, FALSE);
    }
    else
    {
        CString strPatch;
        
        strPatch.LoadString(IDS_PATCH);     
        wsprintf(pszName, "%s %d", (LPCSTR)strPatch, nPatch);
    }
}                                                      

///////////////////////////////////////////////////////////////////////
// Send/receive drums and drumkits

// Send a drumkit to the specified synth
void FAR PASCAL _export GateSetDrumkit(UINT nSynth, LPDRUMKIT lpDrumkit)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;

        Gatekeeper.m_synthArray[nSynth]->SetDrumkit(lpDrumkit);
        Gatekeeper.m_synthArray[nSynth]->CalculatePatchUsages();
        Gatekeeper.m_synthArray[nSynth]->CalculateSampleUsages();
        
        Gatekeeper.m_bIsBusy = FALSE;
    }    
}

// Get a drumkit from the specified synth
void FAR PASCAL _export GateGetDrumkit(UINT nSynth, LPDRUMKIT lpDrumkit)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;

        Gatekeeper.m_synthArray[nSynth]->GetDrumkit(lpDrumkit);

        Gatekeeper.m_bIsBusy = FALSE;
    }    
    else
        memset(lpDrumkit, 0, sizeof(DRUMKIT));
}

// Send a single drum to the specified synth
void FAR PASCAL _export GateSetDrum(UINT nSynth, int nKey, LPDRUM lpDrum)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;

        Gatekeeper.m_synthArray[nSynth]->SetDrum(nKey, lpDrum);
        Gatekeeper.m_synthArray[nSynth]->CalculatePatchUsages();
        Gatekeeper.m_synthArray[nSynth]->CalculateSampleUsages();

        Gatekeeper.m_bIsBusy = FALSE;
    }    
}
    
// Get a single drum from the specified synth    
void FAR PASCAL _export GateGetDrum(UINT nSynth, int nKey, LPDRUM lpDrum)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;

        Gatekeeper.m_synthArray[nSynth]->GetDrum(nKey, lpDrum);

        Gatekeeper.m_bIsBusy = FALSE;
    }    
    else
        memset(lpDrum, 0, sizeof(DRUM));
}

///////////////////////////////////////////////////////////////////////////
// Sample informational functions

// Get the type and size of the specified sample. The size is only valid for
// externally loaded samples (wave files). pdwSampleSize may be NULL, and the 
// call will actually be faster if it is.
int FAR PASCAL _export GateIdentifySample(UINT nSynth, int nSample, 
    LPDWORD pdwSampleSize)
{   
    int nResult = ST_EMPTY;
                                                                     
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           

        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;

        nResult = Gatekeeper.m_synthArray[nSynth]->IdentifySample(nSample, 
            pdwSampleSize);

        Gatekeeper.m_bIsBusy = FALSE;
    }
    else
    {   
        if (pdwSampleSize != NULL)
            *pdwSampleSize = 0;
    }

    return(nResult);
}                                   

// See if the sample can be downloaded to the synth. The function returns both
// a yes/no answer, as well as an error code explaining any failures. Error codes are 
// in WFGATAPI.H.
BOOL FAR PASCAL _export GateIsSampleDownloadable(UINT nSynth, 
    LPCSTR pszFilespec, LPINT pResult)
{                                    
    ASSERT(pResult != NULL);
    
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           
        return(Gatekeeper.m_synthArray[nSynth]->IsSampleDownloadable(pszFilespec, 
            pResult));
    }
    else
    {   
        *pResult = GATE_ERR_NOMEMORY;
        return(FALSE);
    }
}

// See if the Mockingbird sample can be downloaded to the synth. The function returns both
// a yes/no answer, as well as an error code explaining any failures. Error codes are 
// in WFGATAPI.H.
BOOL FAR PASCAL _export GateIsMockingbirdFileDownloadable(UINT nSynth, 
    LPCSTR pszFilespec, LPINT pResult)
{                                    
    ASSERT(pResult != NULL);
    
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           
        return(Gatekeeper.m_synthArray[nSynth]->IsMockingbirdSampleDownloadable(pszFilespec, 
            pResult));
    }
    else
    {   
        *pResult = GATE_ERR_NOMEMORY;
        return(FALSE);
    }
}

// Determine if a specific sample slot in the synth has a ROM sample associated with it.
// This will return TRUE even if the ROM sample has been overwritten.
BOOL FAR PASCAL _export GateIsSampleInternal(UINT nSynth, int nSample)
{                                               
    if (nSynth < Gatekeeper.m_nSynthCount)
        return(Gatekeeper.m_synthArray[nSynth]->IsRomLocation(nSample));
    else
        return(TRUE);
}

// Determine if a sample in a specific location is an original ROM sample
BOOL FAR PASCAL _export GateIsSampleOriginalRom(UINT nSynth, int nSample)
{
    BOOL bResult = FALSE;
    if (nSynth < Gatekeeper.m_nSynthCount)
    {                   
        if (Gatekeeper.m_synthArray[nSynth]->IsRomLocation(nSample))
        {
            char szFilespec[_MAX_PATH];

            Gatekeeper.m_synthArray[nSynth]->GetSampleFilespec(nSample, szFilespec);
            if (szFilespec == Gatekeeper.m_synthArray[nSynth]->m_strInternal)
                bResult = TRUE;
        }
    }
    return(bResult);
}

// Determine if the sample in a specific location is editable. This is
// TRUE for all non-ROM locations. If a ROM sample has been replaced by 
// an external sample, it will return TRUE; otherwise it will return FALSE.
BOOL FAR PASCAL _export GateIsSampleEditable(UINT nSynth, int nSample)
{
    BOOL bResult = FALSE;
    if (nSynth < Gatekeeper.m_nSynthCount)
    {   
        // The sample data is editable if the sample file spec 
        // has a full path that ends in .WAV.
        char szFilespec[_MAX_PATH];

        Gatekeeper.m_synthArray[nSynth]->GetSampleFilespec(nSample, szFilespec);
        strupr(szFilespec);

        if (strstr(szFilespec, ".WAV") != NULL && strstr(szFilespec, "\\") != NULL)
            bResult = TRUE;
    }
    return(bResult);
}

// Find out if a sample is embedded.
BOOL FAR PASCAL _export GateIsSampleEmbedded(UINT nSynth, int nSample)
{
    BOOL bResult = FALSE;

    if (nSynth < Gatekeeper.m_nSynthCount)
        bResult = Gatekeeper.m_synthArray[nSynth]->IsSampleEmbedded(nSample);

    return(bResult);
}   

// Find out if a program is embedded.
BOOL FAR PASCAL _export GateIsProgramEmbedded(UINT nSynth, int nProgram)
{
    BOOL bResult = FALSE;

    if (nSynth < Gatekeeper.m_nSynthCount)
        bResult = Gatekeeper.m_synthArray[nSynth]->IsProgramEmbedded(nProgram);

    return(bResult);
}   

// Find out if a drumkit is embedded.
BOOL FAR PASCAL _export GateIsDrumkitEmbedded(UINT nSynth)
{
    BOOL bResult = FALSE;

    if (nSynth < Gatekeeper.m_nSynthCount)
        bResult = Gatekeeper.m_synthArray[nSynth]->IsDrumkitEmbedded();

    return(bResult);
}   

// The following function is obsolete. Sample names are now always editable
BOOL FAR PASCAL _export GateIsSampleNameEditable(UINT nSynth, int nSample)
{
    return(TRUE);
}    

// Find out how many times the specified sample has been used.
int FAR PASCAL _export GateGetSampleUsageCount(UINT nSynth, int nSample)
{                                               
    if (nSynth < Gatekeeper.m_nSynthCount)
        return(Gatekeeper.m_synthArray[nSynth]->GetSampleUsageCount(nSample));
    else
        return(0);
}

// As above, but if a sample is used by an unused patch, it doesn't count
int FAR PASCAL _export GateGetSampleRelevantUsageCount(UINT nSynth, int nSample)
{                                               
    if (nSynth < Gatekeeper.m_nSynthCount)
        return(Gatekeeper.m_synthArray[nSynth]->GetSampleRelevantUsageCount(nSample));
    else
        return(0);
}

// Locate an unused sample. nFirstSample is the first sample that will be tested.
// If an unused sample is not found above nFirstSample, the program will
// search below nFirstSample. The return value is NO_SAMPLE (-1) if none is found.
// Note that a sample remains unused until a patch, multisample, or alias that 
// uses the sample is actually sent to the Gatekeeper; to find additional unused slots, 
// call this function and pass in the number of the first sample following the last
// unused sample found.
int FAR PASCAL _export GateFindUnusedSample(UINT nSynth, int nFirstSample)
{   
    if (nSynth < Gatekeeper.m_nSynthCount)
        return(Gatekeeper.m_synthArray[nSynth]->FindUnusedSample(nFirstSample)); 
    
    return(NO_SAMPLE);
}

// Get a sample's sample rate. This returns 44100 for internal samples, 0
// for multisamples, or the sample rate for a loaded sample. The sample rate
// of the aliased sample is returned for an alias.
DWORD FAR PASCAL _export GateGetSampleRate(UINT nSynth, int nSample)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           
        return(Gatekeeper.m_synthArray[nSynth]->GetSampleRate(nSample));
    }
    else
        return(44100L);
}

// Get the filespec for the specified sample. If the sample is internal, 
// the word "INTERNAL" is returned. If the sample is a multisample, it returns 
// an empty string. 
void FAR PASCAL _export GateGetSampleFilespec(UINT nSynth, int nSample, LPSTR pszFilespec)
{                                                          
    ASSERT(pszFilespec != NULL);
    
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           

        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;

        Gatekeeper.m_synthArray[nSynth]->GetSampleFilespec(nSample, pszFilespec);

        Gatekeeper.m_bIsBusy = FALSE;
    }
    else
        lstrcpy(pszFilespec, "");
}

// Given one sample of a stereo pair, locate the other channel.
// Returns NO_SAMPLE for multisamples, aliases, and mono samples.
int FAR PASCAL _export GateFindOtherChannel(UINT nSynth, int nSample)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
        return(Gatekeeper.m_synthArray[nSynth]->FindOtherChannel(nSample)); 
    
    return(NO_SAMPLE);
}

// Count the number of .WAV and .MBS files loaded in the synth.
int FAR PASCAL _export GateGetLoadedSampleCount(UINT nSynth)
{
    if (nSynth < Gatekeeper.m_nSynthCount) 
    {   
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        int nSampleCount = 0;
        for (int i = 0; i < NUM_SAMPLES; i++)
        {   
            // If the sample is of type ST_SAMPLE and is not internal, count it
            if (!Gatekeeper.m_synthArray[nSynth]->IsInternalSample(i))
            {
                int nSampleType = Gatekeeper.m_synthArray[nSynth]->IdentifySample(i);            
                if (nSampleType == ST_SAMPLE)
                    nSampleCount++;         
            }
        }
        Gatekeeper.m_bIsBusy = FALSE;

        return(nSampleCount);
    }
    else
        return(0);
}

///////////////////////////////////////////////////////////////////////////
// Send/receive samples, etc.

// Delete the specified sample. This CANNOT delete original ROM samples.
// Be sure to call GateReportFreeMemory() after this to crunch memory!
void FAR PASCAL _export GateDeleteSample(UINT nSynth, int nSample)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());
        if (!GateIsSampleOriginalRom(nSynth, nSample))
        {
            Gatekeeper.WaitUntilNotBusy();
            Gatekeeper.m_bIsBusy = TRUE;

            Gatekeeper.m_synthArray[nSynth]->DeleteSample(nSample);
            Gatekeeper.m_synthArray[nSynth]->CalculateSampleUsages();
            
            Gatekeeper.m_bIsBusy = FALSE;

            // Set the dirty flags for all clients except the current
            for (int i = 0; i < MAX_CLIENTS; i++)
            {                                                                          
                BOOL bDirty = (i != Gatekeeper.m_nCurrentOwner);
                Gatekeeper.m_synthArray[nSynth]->SetSampleDirtyFlag(nSample, i, bDirty);
            }
        }
    }
}

// Set the name of the specified sample.
void FAR PASCAL _export GateSetSampleName(UINT nSynth, int nSample, LPCSTR pszName)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           
        Gatekeeper.m_synthArray[nSynth]->SetSampleName(nSample, pszName);

        // Set the dirty flags for all clients except the current
        for (int i = 0; i < MAX_CLIENTS; i++)
        {                                                                          
            BOOL bDirty = (i != Gatekeeper.m_nCurrentOwner);
            Gatekeeper.m_synthArray[nSynth]->SetSampleDirtyFlag(nSample, i, bDirty);
        }
    }
}

// Get the name (NOT the file name) of the specified sample.
void FAR PASCAL _export GateGetSampleName(UINT nSynth, int nSample, LPSTR pszName)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           
        Gatekeeper.m_synthArray[nSynth]->GetSampleName(nSample, pszName);

        // Clear the dirty flag for the current owner
        Gatekeeper.m_synthArray[nSynth]->SetSampleDirtyFlag(nSample, 
            Gatekeeper.m_nCurrentOwner, FALSE);
    }
    else
    {
        CString strSample;
        
        strSample.LoadString(IDS_SAMPLE);     
        wsprintf(pszName, "%s %d", (LPCSTR)strSample, nSample);
    }
}

// Load a wave file into the synthesizer. The root key from the wave file will
// be used if nRootKey == -1.
BOOL FAR PASCAL _export GateLoadSampleFile(UINT nSynth, int nSample, 
    LPCSTR pszFilespec, int nRootKey)
{
    ASSERT(pszFilespec != NULL);
    
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized()); 

        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;

        BOOL bResult = Gatekeeper.m_synthArray[nSynth]->LoadSampleFile(nSample, 
            pszFilespec, NULL, NULL, nRootKey);

        Gatekeeper.m_bIsBusy = FALSE;

        // Set the dirty flags for all clients except the current
        for (int i = 0; i < MAX_CLIENTS; i++)
        {                                                                          
            BOOL bDirty = (i != Gatekeeper.m_nCurrentOwner);
            Gatekeeper.m_synthArray[nSynth]->SetSampleDirtyFlag(nSample, i, bDirty);
        }
        
        // Set the dirty flags for the other channel, too.
        int nOtherSample = Gatekeeper.m_synthArray[nSynth]->FindOtherChannel(nSample);
        if (nOtherSample != NO_SAMPLE)
            for (i = 0; i < MAX_CLIENTS; i++)
            {                                                                          
                BOOL bDirty = (i != Gatekeeper.m_nCurrentOwner);
                Gatekeeper.m_synthArray[nSynth]->SetSampleDirtyFlag(nOtherSample, i, bDirty);
            }

        return(bResult);
    }
    else
        return(FALSE);
}

// Load a Mockingbird file into the synthesizer. The root key from the wave file will
// be used if nRootKey == -1.
BOOL FAR PASCAL _export GateLoadMockingbirdFile(UINT nSynth, int nSample, 
    LPCSTR pszFilespec, int nRootKey)
{       
    BOOL bResult = FALSE;
    
    ASSERT(pszFilespec != NULL);
    
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized()); 

        // Set the dirty flags for all clients except the current
        for (int i = 0; i < MAX_CLIENTS; i++)
        {                                                                          
            BOOL bDirty = (i != Gatekeeper.m_nCurrentOwner);
            Gatekeeper.m_synthArray[nSynth]->SetSampleDirtyFlag(nSample, i, bDirty);
        }

        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
    
        bResult = Gatekeeper.m_synthArray[nSynth]->LoadMockingbirdFile(nSample, pszFilespec, 
            NULL, NULL, nRootKey);

        Gatekeeper.m_bIsBusy = FALSE;
    }
    
    return(bResult);
}

// Set the sample header of a sample.
void FAR PASCAL _export GateSetSample(UINT nSynth, int nSample, LPSAMPLE lpSample, 
    LPCSTR pszName)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           

        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
    
        Gatekeeper.m_synthArray[nSynth]->SetSampleHeader(nSample, lpSample, pszName);

        Gatekeeper.m_bIsBusy = FALSE;

        // Set the dirty flags for all clients except the current
        for (int i = 0; i < MAX_CLIENTS; i++)
        {                                                                          
            BOOL bDirty = (i != Gatekeeper.m_nCurrentOwner);
            Gatekeeper.m_synthArray[nSynth]->SetSampleDirtyFlag(nSample, i, bDirty);
        }
    }
}

// Get the sample header of a sample.
void FAR PASCAL _export GateGetSample(UINT nSynth, int nSample, LPSAMPLE lpSample, 
    LPSTR pszName)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           

        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
    
        Gatekeeper.m_synthArray[nSynth]->GetSample(nSample, lpSample, pszName);

        Gatekeeper.m_bIsBusy = FALSE;

        // Clear the dirty flag for the current owner
        Gatekeeper.m_synthArray[nSynth]->SetSampleDirtyFlag(nSample, 
            Gatekeeper.m_nCurrentOwner, FALSE);
    }
    else
    {
        CString strSample;
        
        strSample.LoadString(IDS_SAMPLE);     
        wsprintf(pszName, "%s %d", (LPCSTR)strSample, nSample);
        memset(lpSample, 0, sizeof(SAMPLE));
    }
}

// Download a multisample to the synth.
void FAR PASCAL _export GateSetMultisample(UINT nSynth, int nSample, 
    LPMULTISAMPLE lpMultisample, LPCSTR pszName)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           

        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;

        Gatekeeper.m_synthArray[nSynth]->SetMultisample(nSample, lpMultisample, pszName);
        Gatekeeper.m_synthArray[nSynth]->CalculateSampleUsages();

        Gatekeeper.m_bIsBusy = FALSE;

        // Set the dirty flags for all clients except the current
        for (int i = 0; i < MAX_CLIENTS; i++)
        {                                                                          
            BOOL bDirty = (i != Gatekeeper.m_nCurrentOwner);
            Gatekeeper.m_synthArray[nSynth]->SetSampleDirtyFlag(nSample, i, bDirty);
        }
    }
}

// Get a multisample from the synth.
void FAR PASCAL _export GateGetMultisample(UINT nSynth, int nSample, 
    LPMULTISAMPLE lpMultisample, LPSTR pszName)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           

        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->GetMultisample(nSample, lpMultisample, pszName);

        Gatekeeper.m_bIsBusy = FALSE;

        // Clear the dirty flag for the current owner
        Gatekeeper.m_synthArray[nSynth]->SetSampleDirtyFlag(nSample, 
            Gatekeeper.m_nCurrentOwner, FALSE);
    }
    else
    {
        CString strSample;
        
        strSample.LoadString(IDS_SAMPLE);     
        wsprintf(pszName, "%s %d", (LPCSTR)strSample, nSample);
        memset(lpMultisample, 0, sizeof(MULTISAMPLE));
    }
}                                                                                   

// Download an alias to the synth.
void FAR PASCAL _export GateSetAlias(UINT nSynth, int nAlias, LPALIAS lpAlias, 
    LPCSTR pszName)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           

        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->SetAlias(nAlias, lpAlias, pszName);
        Gatekeeper.m_synthArray[nSynth]->CalculateSampleUsages();

        Gatekeeper.m_bIsBusy = FALSE;

        // Set the dirty flags for all clients except the current
        for (int i = 0; i < MAX_CLIENTS; i++)
        {                                                                          
            BOOL bDirty = (i != Gatekeeper.m_nCurrentOwner);
            Gatekeeper.m_synthArray[nSynth]->SetSampleDirtyFlag(nAlias, i, bDirty);
        }
    }
}

// Get an alias from the synth.
void FAR PASCAL _export GateGetAlias(UINT nSynth, int nAlias, LPALIAS lpAlias, 
    LPSTR pszName)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           

        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->GetAlias(nAlias, lpAlias, pszName);

        Gatekeeper.m_bIsBusy = FALSE;

        // Clear the dirty flag for the current owner
        Gatekeeper.m_synthArray[nSynth]->SetSampleDirtyFlag(nAlias, 
            Gatekeeper.m_nCurrentOwner, FALSE);
    }
    else
    {
        CString strSample;
        
        strSample.LoadString(IDS_SAMPLE);     
        wsprintf(pszName, "%s %d", (LPCSTR)strSample, nAlias);
        memset(lpAlias, 0, sizeof(ALIAS));
    }
}

// get the size of a sample, in samples
DWORD FAR PASCAL _export GateGetSampleCount(UINT nSynth, int nSample)
{
    DWORD dwResult = 0;
    
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        ASSERT(Gatekeeper.m_synthArray[nSynth]->IsInitialized());                           

        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        dwResult = Gatekeeper.m_synthArray[nSynth]->GetSampleCount(nSample);

        Gatekeeper.m_bIsBusy = FALSE;
    }
    
    return(dwResult);
}
  

/////////////////////////////////////////////////////////////////////////
// File functions

// Load a bank (.wfb) file
BOOL FAR PASCAL _export GateOpenBankFile(UINT nSynth, LPCSTR szFileSpec)
{                              
    BOOL bResult = FALSE;
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        bResult = Gatekeeper.m_synthArray[nSynth]->OpenBankFile(szFileSpec);

        Gatekeeper.m_bIsBusy = FALSE;
    }
    
    return(bResult);
}

// Save a bank (.wfb) file
BOOL FAR PASCAL _export GateSaveBankFile(UINT nSynth, LPCSTR szFileSpec, 
    BOOL bEmbedded, BOOL bBackup)
{
    BOOL bResult = FALSE;
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        bResult = Gatekeeper.m_synthArray[nSynth]->SaveBankFile(szFileSpec, 
            bEmbedded, bBackup);

        Gatekeeper.m_bIsBusy = FALSE;
    }
    
    return(bResult);
}

// Load a program (.wfp) file
BOOL FAR PASCAL _export GateOpenProgramFile(UINT nSynth, UINT nProgram, 
    LPCSTR szFileSpec)
{
    BOOL bResult = FALSE;
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        bResult = Gatekeeper.m_synthArray[nSynth]->OpenProgramFile(nProgram, 
            szFileSpec);

        Gatekeeper.m_bIsBusy = FALSE;
    }
    
    return(bResult);
}

// Save a program (.wfp) file
BOOL FAR PASCAL _export GateSaveProgramFile(UINT nSynth, UINT nProgram, 
    LPCSTR szFileSpec, BOOL bEmbedded, BOOL bBackup)
{
    BOOL bResult = FALSE;
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        bResult = Gatekeeper.m_synthArray[nSynth]->SaveProgramFile(nProgram, 
            szFileSpec, bEmbedded, bBackup);

        Gatekeeper.m_bIsBusy = FALSE;
    }
    
    return(bResult);
}

// Load a drum kit (.wfd) file
BOOL FAR PASCAL _export GateOpenDrumkitFile(UINT nSynth, LPCSTR szFileSpec)
{
    BOOL bResult = FALSE;
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        bResult = Gatekeeper.m_synthArray[nSynth]->OpenDrumkitFile(szFileSpec);

        Gatekeeper.m_bIsBusy = FALSE;
    }
    
    return(bResult);
}

// Save a drum kit (.wfd) file
BOOL FAR PASCAL _export GateSaveDrumkitFile(UINT nSynth, LPCSTR szFileSpec, 
    BOOL bEmbedded, BOOL bBackup)
{
    BOOL bResult = FALSE;
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        bResult = Gatekeeper.m_synthArray[nSynth]->SaveDrumkitFile(szFileSpec, 
            bEmbedded, bBackup);

        Gatekeeper.m_bIsBusy = FALSE;
    }
    
    return(bResult);
}

// Write the current bank to a set of ROMable files.
BOOL FAR PASCAL _export GateWriteRomFiles(UINT nSynth, LPCSTR szFilespec, LPCSTR szMicrocodeFile,
    UINT nProcessor, UINT nRomSize, UINT nMajorRevision, UINT nMinorRevision, 
    LPCSTR lpszPartNumber, LPCSTR lpszCopyright, BOOL bAddChannelPrograms)
{
    BOOL bResult = FALSE;
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        bResult = Gatekeeper.m_synthArray[nSynth]->WriteRomFiles(szFilespec, szMicrocodeFile,
            nProcessor, nRomSize, nMajorRevision, nMinorRevision, 
            lpszPartNumber, lpszCopyright, bAddChannelPrograms);

        Gatekeeper.m_bIsBusy = FALSE;
    }
    
    return(bResult);
}

// See if the last bank loaded was embedded. Note that loading an embedded
// program or drum kit will force the bank to be embedded as well.  
BOOL FAR PASCAL _export GateIsBankEmbedded(UINT nSynth)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
        return(Gatekeeper.m_synthArray[nSynth]->IsBankEmbedded());
    else
        return(FALSE);
}

// Find the filename and path of the current bank.
void FAR PASCAL _export GateGetBankFilespec(UINT nSynth, LPSTR szFilespec)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
        Gatekeeper.m_synthArray[nSynth]->GetBankFilespec(szFilespec);
    else
        szFilespec[0] = 0;
}
// Get the comment associated with a bank
void FAR PASCAL _export GateGetBankComment(UINT nSynth, LPSTR szComment)
{
    if (nSynth < Gatekeeper.m_nSynthCount)
        lstrcpy(szComment, Gatekeeper.m_synthArray[nSynth]->m_strComment);
    else
        szComment[0] = 0;
}

// Get the comment associated with a bank
void FAR PASCAL _export GateSetBankComment(UINT nSynth, LPCSTR szComment)
{   
    ASSERT(lstrlen(szComment) < MAX_COMMENT);
        
    if (nSynth < Gatekeeper.m_nSynthCount) 
        Gatekeeper.m_synthArray[nSynth]->m_strComment = szComment;
}         

///////////////////////////////////////////////////////////////////////////
// Effect editing functions                                                

// Set the reverb decay time
void FAR PASCAL _export GateSetReverbDecayTime(UINT nSynth, int nDecayTime)
{
    ASSERT(Gatekeeper.m_synthArray[nSynth]->HasEffects());
    
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->SetReverbDecayTime(nDecayTime);

        Gatekeeper.m_bIsBusy = FALSE;
    }
}

// Set the absorption amount
void FAR PASCAL _export GateSetReverbAbsorption(UINT nSynth, int nAbsorption)
{
    ASSERT(Gatekeeper.m_synthArray[nSynth]->HasEffects());
    
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->SetReverbAbsorption(nAbsorption);

        Gatekeeper.m_bIsBusy = FALSE;
    }
}

// Set the room size
void FAR PASCAL _export GateSetReverbRoomSize(UINT nSynth, int nRoomSize)
{
    ASSERT(Gatekeeper.m_synthArray[nSynth]->HasEffects());
    
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->SetReverbRoomSize(nRoomSize);

        Gatekeeper.m_bIsBusy = FALSE;
    }
}

// Set the reverb pre-delay
void FAR PASCAL _export GateSetReverbPredelay(UINT nSynth, int nPredelay)
{
    ASSERT(Gatekeeper.m_synthArray[nSynth]->HasEffects());
    
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->SetReverbPredelay(nPredelay);

        Gatekeeper.m_bIsBusy = FALSE;
    }
}

// Set the reverb mix
void FAR PASCAL _export GateSetReverbMix(UINT nSynth, int nMix)
{
    ASSERT(Gatekeeper.m_synthArray[nSynth]->HasEffects());
    
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->SetReverbMix(nMix);

        Gatekeeper.m_bIsBusy = FALSE;
    }
}

// Set all reverb parameters. This is faster than setting all parameters
// individually
void FAR PASCAL _export GateSetReverbEverything(UINT nSynth, int nMix, 
    int nDecayTime, int nAbsorption, int nRoomSize, int nPredelay)
{
    ASSERT(Gatekeeper.m_synthArray[nSynth]->HasEffects());
    
    if (nSynth < Gatekeeper.m_nSynthCount)
    {
        Gatekeeper.WaitUntilNotBusy();
        Gatekeeper.m_bIsBusy = TRUE;
        
        Gatekeeper.m_synthArray[nSynth]->SetReverbEverything(nMix, 
            nDecayTime, nAbsorption, nRoomSize, nPredelay);

        Gatekeeper.m_bIsBusy = FALSE;
    }
}

// Get the reverb decay time
int FAR PASCAL _export GateGetReverbDecayTime(UINT nSynth)
{
    ASSERT(Gatekeeper.m_synthArray[nSynth]->HasEffects());
    
    if (nSynth < Gatekeeper.m_nSynthCount)
        return(Gatekeeper.m_synthArray[nSynth]->GetReverbDecayTime());
    else
        return(MIN_DECAY);
}

// Get the reverb absorption amount
int FAR PASCAL _export GateGetReverbAbsorption(UINT nSynth)
{
    ASSERT(Gatekeeper.m_synthArray[nSynth]->HasEffects());
    
    if (nSynth < Gatekeeper.m_nSynthCount)
        return(Gatekeeper.m_synthArray[nSynth]->GetReverbAbsorption());
    else
        return(MIN_ABSORPTION);
}

// Get the reverb room size
int FAR PASCAL _export GateGetReverbRoomSize(UINT nSynth)
{
    ASSERT(Gatekeeper.m_synthArray[nSynth]->HasEffects());
    
    if (nSynth < Gatekeeper.m_nSynthCount)
        return(Gatekeeper.m_synthArray[nSynth]->GetReverbRoomSize());
    else
        return(MIN_ROOMSIZE);
}

// Get the reverb pre-delay
int FAR PASCAL _export GateGetReverbPredelay(UINT nSynth)
{
    ASSERT(Gatekeeper.m_synthArray[nSynth]->HasEffects());
    
    if (nSynth < Gatekeeper.m_nSynthCount)
        return(Gatekeeper.m_synthArray[nSynth]->GetReverbPredelay());
    else
        return(MIN_PREDELAY);
}

// Get the reverb mix
int FAR PASCAL _export GateGetReverbMix(UINT nSynth)
{
    ASSERT(Gatekeeper.m_synthArray[nSynth]->HasEffects());
    
    if (nSynth < Gatekeeper.m_nSynthCount)
        return(Gatekeeper.m_synthArray[nSynth]->GetReverbMix());
    else
        return(MIN_MIX);
}  

