// FILES.H
// File header structures, etc. for WaveFront Gatekeeper files
// 3/31/94 J. Johnson                                             
// Copyright  1994 Turtle Beach Systems, Inc. All Rights Reserved

// This file contains all information needed to describe the WaveFront file formats.
// Files are stored, roughly speaking, in the following format:
//       - Header
//       - Program data
//       - Drum data
//       - Patch data
//       - Sample data
//       - Effect data (if applicable)

// Each data element is stored in a structure, defined below, which generally
// duplicates the binary data image sent to the WaveFront synth, with additional 
// information about its name, number, etc.
// Note that, in verson 1.0 of this format, effects storage formats are not defined.

#ifndef FILES_H
#define FILES_H        

const WORD WF_VERSION = 120;  // Current version number (1.2) 
const int  MAX_COMMENT = 64;  // Comment length
const char szBankFile[]    = "Bank";
const char szDrumkitFile[] = "DrumKit";
const char szProgramFile[] = "Program";

const UINT BANKFILE    = 0;
const UINT DRUMKITFILE = 1;
const UINT PROGRAMFILE = 2;

// File header structure. This is used as the file header for all Wavefront data files.
struct WaveFrontFileHeader
{ 
    char  szSynthName[NAME_LENGTH];   // model name of the synthesizer
    char  szFileType[NAME_LENGTH];    // "Bank", "Program", "DrumKit", etc.
    WORD  wVersion;                   // Version times 100; v1.00 = 100, etc.
    WORD  wProgramCount;              // Number of programs in this file
    WORD  wDrumkitCount;              // Number of drumkits in this file
    WORD  wPatchCount;                // Number of patches in this file
    WORD  wSampleCount;               // Number of samples in this file
    WORD  wEffectsCount;              // Number of effects in this file
    DWORD dwProgramOffset;            // Offset to the beginning of program data
    DWORD dwDrumkitOffset;            // Offset to the beginning of drumkit data
    DWORD dwPatchOffset;              // Offset to the beginning of patch data
    DWORD dwSampleOffset;             // Offset to the beginning of sample data
    DWORD dwEffectsOffset;            // Offset to the beginning of effects data
    DWORD dwMemoryRequired;           // Total size in bytes of all samples
    BOOL  bEmbeddedSamples;           // TRUE if samples are stored in this file.
    BOOL  bUnused;                    // Unused
    char  szComment[MAX_COMMENT];     // Comment (primarily for copyright notices)
    BYTE  byPad[62];                  // for expansion (shrinking rapidly!) 
};

// Data structures. Program and patch structures are quite simple; they are just
// the data structs from the synth with names and numbers appended. The drumkit 
// struct is even simpler; it is really here only to make the terminology consistent.
// Sample structures are more complex, as they must deal with more information.  
//
// In files BELOW VERSION 1.20, the WaveFrontSampleInfo struct is used. 
// This is followed by:                           
//
//   - the appropriate structure (SAMPLE, ALIAS, or MULTISAMPLE) as sent
//     to the synth, AND for samples only 
//   - the full path to the sample file (_MAX_PATH characters in length). 
//     If the sample is internal, this will be "INTERNAL".
//     If the sample is an embedded file, this will be "EMBEDDED".
//     AND, if the file is embedded (and the sample is not internal) - 
//   - an image of the wave or MBS file. This is intended to be saved in a 
//     temporary disk file before it is dumped to the synth.
//
// For files of version 1.2 and up, the WavefrontExtendedSampleInfo struct is used. 
//     This is followed by the same types of info as in earlier versions, EXCEPT that
//     embedded data is raw sample data; no file headers are used.
//

struct WaveFrontProgram : public PROGRAM
{
    int  nNumber;             // The original location of the program
    char szName[NAME_LENGTH]; // The sample name
};

struct WaveFrontPatch : public PATCH
{
    int  nNumber;             // The original location of the patch
    char szName[NAME_LENGTH]; // The patch name
};

struct WaveFrontDrumkit : public DRUMKIT
{
};

// Struct for sample info. This is only used by versions 1.0 and 1.1 of the program
struct WaveFrontSampleInfo
{
    DWORD dwSize;               // Size of this sample, including this struct and any following data
    int   nSampleType;          // ST_SAMPLE, etc. constants from WFSYNTH.H
    int   nNumber;              // The original location of the sample
    char  szName[NAME_LENGTH];  // Name of the sample
};  

// Struct for extended sample info. This is for versions 1.2 and up. Note the similarity
// to the above structure.
struct WaveFrontExtendedSampleInfo
{
    DWORD dwSize;               // Size of this sample, including this struct and any following data
    int   nSampleType;          // ST_SAMPLE, etc. constants from WFSYNTH.H
    int   nNumber;              // The original location of the sample
    char  szName[NAME_LENGTH];  // Name of the sample
    DWORD dwSampleRate;         // Sample rate in Hz
    DWORD dwSizeInBytes;        // Size of the sample data itself, in bytes
    DWORD dwSizeInSamples;      // Number of samples in the file
    UINT  nChannel;             // CH_ constants from WAVE.H
    BYTE  byUnused[62];         // For expansion
};  

#endif          

